#include "InvCompAlgorithm.h"

// .
CInvCompRegistrationAlgorithm::CInvCompRegistrationAlgorithm()
{
  //     .
  m_pImgWarpedI2 = NULL;
  m_pImgGradI1x = NULL;
  m_pImgGradI1y = NULL;
  m_pMatInverseHessianF = NULL;
  m_pMatGradF = NULL;
  m_pMatA = NULL;  
  m_pMatV = NULL;  
  m_pMatPrevParams = NULL;  
  m_pMatNewParams = NULL;  
}

// .
CInvCompRegistrationAlgorithm::~CInvCompRegistrationAlgorithm()
{
  //   .
  FreeResources();
}

bool CInvCompRegistrationAlgorithm::RegisterImages(
    IplImage* pImgI1,        
    IplImage* pImgConfMap,   
    IplImage* pImgI2,        
    CPixelMappingModel* pPixelMapping, 
    float fAccuracy,  
    int nMaxIters,     
    int nMaxOscillatingIters, 
    int* pnActualIters, 
    float fInnerAccuracy, 
    int nInnerMaxIters 
   )
{ 
  //   ( )  .
  if(pnActualIters)
    *pnActualIters = 0;

  CriteriaContext context;

  //  .
  CvSize ImageSize = cvSize(pImgI1->width, pImgI1->height);
  if(!AllocResources(ImageSize, pPixelMapping))
    return false;

  //   .
  if(!Precompute(pImgI1, pImgConfMap, pPixelMapping))
    return false;

  //   .
  int l;
  for(l=0; ; l++)
  {     
    //   I2.
    WarpImageI2(pImgI2, pImgConfMap, pPixelMapping);
          
    //    
    //     .
    cvSet(m_pMatPrevParams, cvScalar(0));

    //   .
    float F = 0;

    //   .
    int k;
    for(k = 0; ; k++)
    {        
      //      .
      F = ApproximateParameterIncrement(pImgI1, pImgConfMap, 
        pPixelMapping, m_pMatPrevParams, m_pMatNewParams);
      if(F<0)
        return false;
    
      //     .
      CRITERIA_TYPE InnerLoopReady = IsInnerLoopReady(m_pMatPrevParams, 
        m_pMatNewParams, fInnerAccuracy, k, nInnerMaxIters);
        
      //      .
      cvCopy(m_pMatNewParams, m_pMatPrevParams);        

      //   ,    .
      if(InnerLoopReady!=NONE)
        break;
    }
    
    //     .
    if(!pPixelMapping->Update(m_pMatNewParams))
      return false;

    //      .
    if(IsReady(m_pMatNewParams, fAccuracy, l, nMaxIters, 
      F, nMaxOscillatingIters, context ))
      break;
  }

  //       .
  if(pnActualIters)
    *pnActualIters = l;
  
  return true;
}

bool 
CInvCompRegistrationAlgorithm::AllocResources(
  CvSize ImageSize, 
  CPixelMappingModel* pPixelMapping
 )
{
  //   .
  if(ImageSize.width<=0 ||
     ImageSize.height<=0 ||
     pPixelMapping==NULL)
  {    
    assert(0); //   
    return false;
  }

  //     , 
  //  .
  if(m_pImgWarpedI2==NULL ||
     m_pImgWarpedI2->width!=ImageSize.width ||
     m_pImgWarpedI2->height!=ImageSize.height)
  {
    if(m_pImgWarpedI2!=NULL)
    {
      //   .
      cvReleaseImage(&m_pImgWarpedI2);
      cvReleaseImage(&m_pImgGradI1x);
      cvReleaseImage(&m_pImgGradI1y);
    }

    //  ,   
    //   I2.
    m_pImgWarpedI2 = cvCreateImage(ImageSize, IPL_DEPTH_8U, 1);

    //     .
    // dI1/dx
    m_pImgGradI1x = cvCreateImage(ImageSize, IPL_DEPTH_32F, 1); 
    // dI1/dy
    m_pImgGradI1y = cvCreateImage(ImageSize, IPL_DEPTH_32F, 1); 

    int nParamCount = pPixelMapping->GetParamCount();
    if(m_pMatInverseHessianF==NULL ||
       m_pMatInverseHessianF->rows!=nParamCount)
    {
      //     .
      size_t i;
      for(i=0; i<m_apImgSteepestDescent.size(); i++)
       cvReleaseImage(&m_apImgSteepestDescent[i]);            
      m_apImgSteepestDescent.clear();
      
      //    .
      for(i=0; i<(size_t)nParamCount; i++)
      {
        IplImage* pImg = cvCreateImage(ImageSize, IPL_DEPTH_32F, 1);
        m_apImgSteepestDescent.push_back(pImg);
      }
    }
  }

  /* C . */
  int nParamCount = pPixelMapping->GetParamCount();
  if(m_pMatGradF==NULL ||
    m_pMatGradF->rows!=nParamCount)
  { 
    if(m_pMatGradF!=NULL)
    {
      cvReleaseMat(&m_pMatGradF);
      cvReleaseMat(&m_pMatInverseHessianF);
      cvReleaseMat(&m_pMatA);      
      cvReleaseMat(&m_pMatV);
      cvReleaseMat(&m_pMatPrevParams);
      cvReleaseMat(&m_pMatNewParams);
    }

    //   m x 1     
    //    .
    m_pMatGradF = cvCreateMat(nParamCount, 1, CV_32FC1);
    
    //  m x m     .
    m_pMatInverseHessianF = cvCreateMat(nParamCount, nParamCount, CV_32FC1);
    
    //    m x m.
    m_pMatA = cvCreateMat(nParamCount, nParamCount, CV_32FC1);   

    //    m x 1.
    m_pMatV = cvCreateMat(nParamCount, 1, CV_32FC1);

    //      m x 1.
    m_pMatPrevParams = cvCreateMat(nParamCount, 1, CV_32FC1);

    //      m x 1.
    m_pMatNewParams = cvCreateMat(nParamCount, 1, CV_32FC1);
  }
  
  return true;
}

void 
CInvCompRegistrationAlgorithm::FreeResources()
{
  //   .
  cvReleaseImage(&m_pImgWarpedI2);
  cvReleaseImage(&m_pImgGradI1x);
  cvReleaseImage(&m_pImgGradI1y);

  //    .
  size_t i;
  for(i=0; i<m_apImgSteepestDescent.size(); i++)
    cvReleaseImage(&m_apImgSteepestDescent[i]);            
  m_apImgSteepestDescent.clear();
    
  //   .  
  cvReleaseMat(&m_pMatGradF);
  cvReleaseMat(&m_pMatInverseHessianF);
  cvReleaseMat(&m_pMatA);  
  cvReleaseMat(&m_pMatV);
  cvReleaseMat(&m_pMatPrevParams);
  cvReleaseMat(&m_pMatNewParams);
}

bool 
CInvCompRegistrationAlgorithm::Precompute(
  IplImage* pImgI1,  
  IplImage* pImgConfMap, 
  CPixelMappingModel* pPixelMapping 
 )
{
  //   .
  if(pImgI1==NULL ||
     pImgConfMap==NULL ||
     pPixelMapping==NULL)
  {
    assert(0); //   .
    return false;
  }

  //  : 
  //   a)  (dI1/dx, dI1/dy)  I1;
  //   )   .
  //   )     .
  //    ,   
  // .

  int nParamCount = pPixelMapping->GetParamCount();

  //     .
  CvMat* pMatJacobian = cvCreateMat(2, 
    nParamCount, CV_32FC1);

  //    I1.

  //  dI1/dx.
  cvSobel(pImgI1, m_pImgGradI1x, 1, 0, 3); 
  // ,    1/8.
  cvCvtScale(m_pImgGradI1x, m_pImgGradI1x, 0.125); 

  //  dI1/dy. 
  cvSobel(pImgI1, m_pImgGradI1y, 0, 1, 3);
  // ,    1/8.
  cvCvtScale(m_pImgGradI1y, m_pImgGradI1y, 0.125); 
  
  //       .

  //        .
  cvSet(m_pMatA, cvScalar(0)); 
    
  //      .
  int x, y;
  for(x=0; x<pImgI1->width; x++)
  {
    for(y=0; y<pImgI1->height; y++)
    { 
      //    .
      float fWeight = (float)CV_IMAGE_ELEM(pImgConfMap, uchar, y, x);
      if(fWeight==0)
        continue;

      //      .
      CvPoint2D32f p = {(float)x, (float)y};
      if(false==pPixelMapping->CalcJacobianMat(pMatJacobian, &p))
      {
        //    !
        cvReleaseMat(&pMatJacobian);
        return false;
      }

      //      .
      float fGradI1x = CV_IMAGE_ELEM(m_pImgGradI1x, float, y, x);
      float fGradI1y = CV_IMAGE_ELEM(m_pImgGradI1y, float, y, x);
      
      //    " ".      
      int k;
      for(k=0; k<nParamCount; k++)
      {
        CV_IMAGE_ELEM(m_apImgSteepestDescent[k], float, y, x) = 
          fGradI1x * CV_MAT_ELEM(*pMatJacobian, float, 0, k) +
          fGradI1y * CV_MAT_ELEM(*pMatJacobian, float, 1, k);
      }
            
      //     .
      int i, j;
      for(i=0; i<nParamCount; i++)
      {
        for(j=0; j<nParamCount; j++)
        {
          CV_MAT_ELEM(*m_pMatA, float, j, i) += fWeight*
            CV_IMAGE_ELEM(m_apImgSteepestDescent[j], float, y, x)*
            CV_IMAGE_ELEM(m_apImgSteepestDescent[i], float, y, x);
        }
      }
    }
  }

  // ,     2.
  cvCvtScale(m_pMatA, m_pMatA, 2);  
  
  //    .
  double fDet = cvInvert(m_pMatA, m_pMatInverseHessianF);
  if(fDet==0)
  {
    //  !
    cvReleaseMat(&pMatJacobian);
    return false;
  }

  cvReleaseMat(&pMatJacobian);
  return true;
}

float 
CInvCompRegistrationAlgorithm::ApproximateParameterIncrement(
  IplImage* pImgI1,      
  IplImage* pImgConfMap, 
  CPixelMappingModel* pPixelMapping, 
  CvMat* pMatPrevParams, 
  CvMat* pMatNewParams   
 )
{
  //   .
  if(pImgI1==NULL ||
     pImgConfMap==NULL ||
     pPixelMapping==NULL ||
     pMatPrevParams==NULL ||
     pMatNewParams==NULL)
  {
    //  
    return -1;
  }
  
  //   ,    
  //  .
  int nParamCount = pPixelMapping->GetParamCount();

  //     .
  float F = 0;

  //    .

  //     .
  cvSet(m_pMatGradF, cvScalar(0)); 
    
  //   .
  int x, y;
  for(x=0; x<pImgI1->width; x++)
  {
    for(y=0; y<pImgI1->height; y++)
    { 
      //    .
      float fWeight = (float)CV_IMAGE_ELEM(pImgConfMap, uchar, y, x);
      if(fWeight==0)
        continue;

      //   I1(x)-WarpedI2(x).
      float fSub = (float) (CV_IMAGE_ELEM(pImgI1, uchar, y, x) - 
                      CV_IMAGE_ELEM(m_pImgWarpedI2, uchar, y, x));
            
      float fTerm = 0;
      int k;
      for(k=0; k<nParamCount; k++)
      {        
        fTerm += CV_MAT_ELEM(*pMatPrevParams, float, k, 0)*
          CV_IMAGE_ELEM(m_apImgSteepestDescent[k], float, y, x);
      }
      fTerm += fSub;

      //      .
      F += fWeight*fTerm*fTerm;

      //     .
      for(k=0; k<nParamCount; k++)
      {
        CV_MAT_ELEM(*m_pMatGradF, float, k, 0) +=
          CV_IMAGE_ELEM(m_apImgSteepestDescent[k], float, y, x)*fTerm*fWeight;
      }
    }
  }
        
  //      2.
  cvCvtScale(m_pMatGradF, m_pMatGradF, 2);
    
  //     Delta_p.
  cvGEMM(m_pMatInverseHessianF, m_pMatGradF, 1, 0, 0, m_pMatV);  
  cvScaleAdd(m_pMatV, cvScalar(-1), pMatPrevParams, pMatNewParams);

  //    .
  return F;
}

float 
CInvCompRegistrationAlgorithm::InterpolatePixelIntensity(
  IplImage* pImg, 
  float x,        
  float y         
  )
{
  //       (xi;yi).
  int xi = cvFloor(x);
  int yi = cvFloor(y);

  //    .
  float k1 = x-xi; 
  float k2 = y-yi;

  // ,      
  //   .
  int f1 = xi<pImg->width-1;  
  int f2 = yi<pImg->height-1; 

  uchar* row1 = &CV_IMAGE_ELEM(pImg, uchar, yi, xi);
  uchar* row2 = &CV_IMAGE_ELEM(pImg, uchar, yi+1, xi);
                
  //   .
  float fInterpolatedValue = (1.0f-k1)*(1.0f-k2)*(float)row1[0] +
     (f1 ? ( k1*(1.0f-k2)*(float)row1[1] ):0) +
     (f2 ? ( (1.0f-k1)*k2*(float)row2[0] ):0) +                        
     ((f1 && f2) ? ( k1*k2*(float)row2[1] ):0) ;

  return fInterpolatedValue;
}

IplImage* CInvCompRegistrationAlgorithm::WarpImageI2(
    IplImage* pImgI2, 
    IplImage* pImgConfMap, 
    CPixelMappingModel* pPixelMapping 
   )
{
  //   WarpedI2 .
  cvSet(m_pImgWarpedI2, cvScalar(0));

  CvPoint2D32f SrcPixel;
  CvPoint2D32f DstPixel;

  //    I2.
  int x, y;
  for(x=0; x<pImgI2->width; x++)
  {
    for(y=0; y<pImgI2->height; y++)
    {
      //    .
      float fWeight = CV_IMAGE_ELEM(pImgConfMap, uchar, y, x);
      if(fWeight==0)
        continue; //     .

      SrcPixel.x = (float)x;
      SrcPixel.y = (float)y;
      
      //        
      // DstPixel,   SrcPixel.
      bool bMap = pPixelMapping->MapPixel(&SrcPixel, &DstPixel);

      // ,     .
      if( bMap &&
          DstPixel.x>=0 && DstPixel.x<m_pImgWarpedI2->width &&
          DstPixel.y>=0 && DstPixel.y<m_pImgWarpedI2->height )
      {
        //     .
        float fInterpolatedVal = 
          InterpolatePixelIntensity(pImgI2, DstPixel.x, DstPixel.y);

        //   .
        CV_IMAGE_ELEM(m_pImgWarpedI2, uchar, y, x) = 
          (uchar)fInterpolatedVal;
      }
      else
      {
        CV_IMAGE_ELEM(m_pImgWarpedI2, uchar, y, x) = 0;
      }      
    }
  }

  //   .
  return m_pImgWarpedI2;
}

CRITERIA_TYPE 
CInvCompRegistrationAlgorithm::IsInnerLoopReady(
  CvMat* pMatPrevParams, 
  CvMat* pMatNewParams,  
  float fAccuracy,   
  int nCurrentIter,  
  int nMaxIters    
 )
{
  // ,     - .
  if(nCurrentIter+1>=nMaxIters)
  {
    //  .  .
    return MAX_ITERS_REACHED;
  }

  //    (.    ).
  float fMax = 0;
  int i;
  for(i=0; i<pMatNewParams->rows; i++)
  {
    float fDiff = CV_MAT_ELEM(*pMatNewParams, float, i, 0) - 
                  CV_MAT_ELEM(*pMatPrevParams, float, i, 0);

    if(i==0 || fabs(fDiff)>fMax)
      fMax = fabs(fDiff);
  }

  if(fMax<fAccuracy)
  {       
    //   .
    return ACCURACY_REACHED;
  }

  //      .
  return NONE;
}

CRITERIA_TYPE 
CInvCompRegistrationAlgorithm::IsReady(
    CvMat* pMatParams, 
    float fAccuracy,   
    int nCurrentIter,  
    int nMaxIters,    
    float fObjFuncVal, 
    int nMaxOscillatingIters,   
    CriteriaContext& context  
   )
{
  // ,     
  //    .
  if(context.m_fMinObjFuncVal<0)
  {
    context.m_fMinObjFuncVal = fObjFuncVal;
    context.m_nBadIters = 0;
  }
  else
  {    
    if(fObjFuncVal<context.m_fMinObjFuncVal)
    {
      context.m_fMinObjFuncVal = fObjFuncVal;
      context.m_nBadIters = 0;    
    }
    else
    {
      //    .     
      context.m_nBadIters++;    
    }
  }

  if(context.m_nBadIters>=nMaxOscillatingIters)
  {
    //    
    //    .
    return OSCILLATION;
  }

  // ,     - .
  if(nCurrentIter+1>=nMaxIters)
  {
    //  .  .
    return MAX_ITERS_REACHED;
  }

  // ,     .

  //    (.    ).
  float fMax = 0;
  int i;
  for(i=0; i<pMatParams->rows; i++)
  {
    float fElem = CV_MAT_ELEM(*pMatParams, float, i, 0);

    if(i==0 || fabs(fElem)>fMax)
      fMax = fabs(fElem);
  }

  if(fMax<fAccuracy)
  {     
    //   .
    return ACCURACY_REACHED;
  }

  //      .
  return NONE;
};

